"""
Generally useful predicates to be used when building and configuring processors.

This module contains both predicates and predicate factories ( functions that take configuration
parameters and return predicates).

A predicate name is terminated in `_p`  (e.g. `true_p`).

A predicate factory is terminated in '_pg' (e.g. 'not_pf').

A predicate factory generates predicates.

For example, although the `false_p` is already defined, one could define a functional `false_p` predicate
by using the `not_pf` and the `true_p` predicate like this:

    my_true_p = not_pf(false_p)

"""
from typing import Callable, Any, Sequence

from .utils.util import is_empty
from .xml_util import is_tag


def true_p(elm):
    """
    Returns True regardless of the argument

    >>> true_p(None)
    True
    >>> true_p(1)
    True
    """
    return True


def false_p(elm):
    """
    Returns False regardless of the argument

    >>> false_p(None)
    False
    >>> false_p(1)
    False
    """
    return False


def not_pf(predicate: Callable[[Any], bool]):
    """
    Negates the predicate

    * **predicate**: predicate to be tested
    * **return**: a predicate that is the negation of the passed predicate

    >>> p = not_pf(true_p)
    >>> p(1)
    False
    >>> p = not_pf(false_p)
    >>> p(1)
    True

    """

    def internal(elm):
        return not predicate(elm)

    return internal


def or_pf(*args: Callable[[Any], bool]):
    """
    Does a logical or of the results of the predicates, it shortcuts the processing (returns True as
    soon as one predicate succeeds). If no predicate is passed it returns False

    * **args**: the predicates
    * **return**: a logical or of the results of the predicates applied on the passed elment

    >>> p = or_pf(true_p, true_p)
    >>> p(1)
    True
    >>> p = or_pf(true_p, false_p)
    >>> p(1)
    True
    >>> p = or_pf(false_p, true_p)
    >>> p(1)
    True
    >>> p = or_pf(false_p, false_p)
    >>> p(1)
    False
    >>> p = or_pf()
    >>> p(1)
    False

    """

    def internal(elm):
        for predicate in args:
            if predicate(elm):
                return True
        return False

    return internal


def and_pf(*args: Callable[[Any], bool]):
    """
    Does a logical and of the results of the predicates, it shortcuts the processing (returns False as
    soon as one predicate fails). If no predicate is passed it returns True

    * **args**: the predicates
    * **return**: a logical or of the results of the predicates applied on the passed elment

    >>> p = and_pf(true_p, true_p)
    >>> p(1)
    True
    >>> p = and_pf(true_p, false_p)
    >>> p(1)
    False
    >>> p = and_pf(false_p, true_p)
    >>> p(1)
    False
    >>> p = and_pf(false_p, false_p)
    >>> p(1)
    False
    >>> p = and_pf()
    >>> p(1)
    True

    """

    def internal(elm):
        for predicate in args:
            if not predicate(elm):
                return False
        return True

    return internal


def is_tag_p(elm):
    """
    Tries to check if the element is a tag
    It checks by verifying that the element has a  not None name that is not the string '[doc]'.

    * **elm**: the element to be checked
    * **return**: True if the element looks like a tag
    """
    if elm is not None and elm.name is not None and elm.name != "[doc]":
        return True
    return False


def is_soup_p(elm):
    """
    Tries to check if the element is a tag
    It checks by verifying that the element has a  not None name that is the string '[doc]'.

    * **elm**: the element to be checked
    * **return**: True if the element looks like a BeautifulSoup object
    """
    return elm is not None and elm.name is not None and elm.name == "[doc]"


def is_tag_or_soup_p(elm):
    """
    Tries to check if the element is either a Tag or a BeautifulSoup element
    It checks by verifying that the element has a non None name.

    * **elm**: the element
    * **return**: True if the element is either a Tag or a BeautifulSoup element
    """
    return elm is not None and elm.name is not None


def is_string_p(elm):
    """
    Tries to check if the element is a NavigableString
    It checks by verifying that the element has a None name

    * **elm**: the element
    * **return**: True if the element looks like a NavigableString
    """
    return elm is not None and elm.name is None


def has_name_pf(name_p, ignore_case=True):
    """
    Predicate factory, returns true if the element name matches the pred parameter

    * **name_p**: something that can be converted into a string compare predicate
    * **ignore_case**: should the comparison be case sensitive (default: ignore case)
    * **return**:

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><span>s1</span><SPAN>s2</SPAN></html>'
    >>> s = BeautifulSoup(doc, 'xml')
    >>> e1 = s.html.span
    >>> e2 = s.html.SPAN
    >>> p1 = has_name_pf('span')
    >>> p2 = has_name_pf('span', False)
    >>> p3 = has_name_pf('x')
    >>> p4 = has_name_pf('X', False)
    >>> p1(e1)
    True
    >>> p2(e1)
    True
    >>> p1(e2)
    True
    >>> p2(e2)
    False
    >>> p3(e1)
    False
    >>> p4(e1)
    False

    """
    # pred is a string
    pred = to_string_compare_predicate_pf(name_p, ignore_case)

    def internal(elm):
        if not is_tag_or_soup_p(elm):
            return False
        return pred(elm.name)

    return internal


def has_attribute_pf(attr_p):
    """
    Predicate factory that checks that the element has a particular attribute

    * **attr_p**: something that can be converted to a string compare predicate
    * **return**: a predicate that checks if the current element has the required attribute

    >>> from bs4 import BeautifulSoup
    >>> doc = ('<html><span id="s1" class="c1 c2 c3" style="margin: 3">s1</span>'+
    ... '<div data-x="abc" id="d1">d1</div></html>')
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> e1 = s.html.span
    >>> e2 = s.html.div
    >>> p1 = has_attribute_pf("id")
    >>> p1(e1)
    True
    >>> p1(e2)
    True
    >>> p2 = has_attribute_pf(["class", "data-w"])
    >>> p2(e1)
    True
    >>> p2(e2)
    False
    >>> p3 = has_attribute_pf( lambda x: x in ["data-x","data-xx"])
    >>> p3(e1)
    False
    >>> p3(e2)
    True

    """
    pred = to_string_compare_predicate_pf(attr_p, ignore_case=False)

    def internal(elm):
        if not is_tag(elm):
            return False
        for attr_name in elm.attrs.keys():
            if pred(attr_name):
                return True
        return False

    return internal


def has_attribute_value_pf(attr_p, value_p, ignore_case_value=False):
    """
    Predicate factory that checks that the element has an attribute that passes
    the predicate attr_p(<attribute_name>) and attr_v(<attribute_value)

    * **attr_p**: something that can be converted to a string compare predicate
    * **value_p**:  something that can be converted to a string compare predicate
    * **return**: a predicate that checks if the current element has an attribute with a specific value

    >>> from bs4 import BeautifulSoup
    >>> doc = ('<html><span id="s1" class="c1 c2 c3" style="margin: 3">s1</span>'+
    ... '<div data-x="abc" id="d1">d1</div></html>')
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> e1 = s.html.span
    >>> e2 = s.html.div
    >>> p1 = has_attribute_value_pf("id", "s1")
    >>> p1(e1)
    True
    >>> p1(e2)
    False
    >>> p2 = has_attribute_value_pf(["id", "data-w"], ["s1", "d1"])
    >>> p2(e1)
    True
    >>> p2(e2)
    True
    >>> p3 = has_attribute_value_pf( lambda x: x in ["data-x","data-xx"], "abc")
    >>> p3(e1)
    False
    >>> p3(e2)
    True

    """
    pred_a = to_string_compare_predicate_pf(attr_p, ignore_case=False)
    pred_v = to_string_compare_predicate_pf(value_p, ignore_case=ignore_case_value)

    def internal(elm):
        if not is_tag(elm):
            return False
        for attr_name, attr_value in elm.attrs.items():
            if pred_a(attr_name) and pred_v(attr_value):
                return True
        return False

    return internal


def has_class_pf(class_p):
    """
    A predicate factory that checks that an element has the desired class

    * **class_p**: something that can be converted to a string predicate
    * **return**: a predicate that returns True for elements that have classes that satisfy class_p

    >>> from bs4 import BeautifulSoup
    >>> doc = ('<html><span class="c1 c2 c3" >s1</span><div id="d1">d1</div></html>')
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> e1 = s.html.span
    >>> e2 = s.html.div
    >>> p = has_class_pf("c2")
    >>> p(e1)
    True
    >>> p(e2)
    False
    >>> p = has_class_pf(["c5","c2"])
    >>> p(e1)
    True
    >>> p(e2)
    False
    >>> p = has_class_pf(["c5","c6"])
    >>> p(e1)
    False

    """
    pred = to_string_compare_predicate_pf(class_p)

    def internal(elm):
        if not is_tag(elm):
            return False

        clss = elm.attrs.get("class", [""])
        for cls in clss:
            if pred(cls):
                return True
        return False

    return internal


def has_children_of_type_pf(name_p, ignore_case=True):
    """
    Creates a predicate that checks if the immediate descendents of the element have a name that
    satisfies name_p

    * **name_p**: something that can be converted into a string predicate
    * **return**: true if the current element has children that satisfy the predicate

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><span>s1</span><div>d1 <p>hello</p></div></html>'
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> e = s.html
    >>> pred = has_children_of_type_pf("span")
    >>> pred(e)
    True
    >>> pred = has_children_of_type_pf("div")
    >>> pred(e)
    True
    >>> pred = has_children_of_type_pf("p")
    >>> pred(e)
    False
    >>> pred = has_children_of_type_pf("a")
    >>> pred(e)
    False

    """
    pred = to_string_compare_predicate_pf(name_p, ignore_case)

    def internal(elm):
        if not is_tag_or_soup_p(elm):
            return False

        for child in elm.children:
            if is_tag_or_soup_p(child) and pred(child.name):
                return True
        return False

    return internal


def has_descendents_of_type_pf(name_p, ignore_case=True):
    """
    Creates a predicate that checks if the descendents of the element have a name that
    satisfies name_p (it will go deep looking for the elements).

    * **name_p**: something that can be converted into a string predicate
    * **return**: true if the current element has children that satisfy the predicate

    >>> from bs4 import BeautifulSoup
    >>> doc = ('<html><span id="s1" class="c1 c2 c3" style="margin: 3">s1</span>'+
    ... '<div data-x="abc" id="d1">d1</div></html>')
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> e1 = s.html.span
    >>> e2 = s.html.div

    """
    pred = to_string_compare_predicate_pf(name_p, ignore_case)

    def internal(elm):
        if not is_tag_or_soup_p(elm):
            return False

        for child in elm.children:
            if is_tag_or_soup_p(child) and pred(child.name):
                return True
            if internal(child):
                return True
        return False

    return internal


def is_empty_p(elm):
    """
    Returns true if the elm  is an empty Navigable string or an elm with at most empty navigatable string chidlren

    * **elm**: a Beautiful soup element
    * **return**: True if the element is an empty string

    >>> from bs4 import BeautifulSoup as bs
    >>> doc = bs("<html><div>   <span>hello</span> <p></p></div></html>", "html.parser")
    >>> div = doc.html.div
    >>> span = doc.html.div.span
    >>> p = doc.html.div.p
    >>> is_empty_p(div)
    False
    >>> is_empty_p(span)
    False
    >>> is_empty_p(p)
    False
    >>> children = list(div.children)
    >>> children[0]
    ' '
    >>> is_empty_p(children[0])
    True
    >>> children[1]
    <span>hello</span>
    >>> is_empty_p(children[1])
    False
    >>> children[2]
    ' '
    >>> is_empty_p(children[2])
    True
    >>> children[3]
    <p></p>
    >>> is_empty_p(children[3])
    False
    """
    if elm is None:
        return True
    if is_string_p(elm):
        return is_empty(elm)

    for child in elm.children:
        if is_tag(child):
            return False
        if not is_empty(child):
            return False

    return True


def to_string_compare_predicate_pf(pred, ignore_case=True):
    """
    Turns the passed parameter into a string compare predicate

    * **pred**: can be one of:
        - a string ( will compare to the string) (e.g. `to_string_compare_predicate_pf( 'div')` ... predicate
            checking if the passed argument is `"div"`
        - a list,tuple,set,frozenset : will create a predicate that checks if the tag is one of the passed tags
            (e.g.  `to_string_compare_predicate_pf(['div','span','p')` checks if the passed argument is one of `[div,
            span, p]`)
        - a predicate `Callable[[string,bool],bool]`, checks if the passed argument has a name that satisfies the
        predicate (where the second parameter is the ignore_case param from the factory).
    * **ignore_case**: should the comparison be case sensitive (default: ignore case)
    * **return**: a predicate that expects a string like object

    >>> e1 = 'span'
    >>> e2 = 'SPAN'
    >>> e3 = "ul"
    >>> p1 = to_string_compare_predicate_pf('span')
    >>> p2 = to_string_compare_predicate_pf('span', False)
    >>> p3 = to_string_compare_predicate_pf('x')
    >>> p4 = to_string_compare_predicate_pf('X', False)
    >>> p1(e1)
    True
    >>> p2(e1)
    True
    >>> p1(e2)
    True
    >>> p2(e2)
    False
    >>> p3(e1)
    False
    >>> p4(e1)
    False

    >>> p5 = to_string_compare_predicate_pf(['i','span','a'])
    >>> p6 = to_string_compare_predicate_pf(['i','span','a'], False)
    >>> p7 = to_string_compare_predicate_pf(['i','x','a'])
    >>> p5(e1)
    True
    >>> p5(e2)
    True
    >>> p6(e1)
    True
    >>> p6(e2)
    False
    >>> p7(e1)
    False
    >>> p8 = to_string_compare_predicate_pf({'i','span','a'})
    >>> p8(e1)
    True
    >>> p9 = to_string_compare_predicate_pf(('i','span','a'))
    >>> p9(e1)
    True
    >>> p10 = to_string_compare_predicate_pf(frozenset(['i','span','a']))
    >>> p10(e1)
    True
    >>> p10(e2)
    True
    >>> def pred(elm_name):
    ...    if elm_name == 'span' or elm_name == 'x':
    ...        return True
    ...    return False
    >>> p11 = to_string_compare_predicate_pf(pred)
    >>> p12 = to_string_compare_predicate_pf(pred, ignore_case=False)
    >>> p11(e1)
    True
    >>> p11(e2)
    True
    >>> p11(e3)
    False
    >>> p12(e1)
    True
    >>> p12(e2)
    False

    """
    if pred is None:
        return False
    # pred is a string
    if isinstance(pred, str):
        name = pred.lower() if ignore_case else pred

        def case_sensitive_compare(elm_name):
            return elm_name == name

    # pred is a list like object
    elif isinstance(pred, (tuple, list, set, frozenset)):
        names = frozenset([name.lower() for name in pred]) if ignore_case else frozenset(pred)

        def case_sensitive_compare(elm_name):
            return elm_name in names
    # pred must be a predicate, use it as is
    else:
        case_sensitive_compare = pred

    if ignore_case:
        def ret_val(elm_name):
            return case_sensitive_compare(elm_name.lower())
    else:
        ret_val = case_sensitive_compare

    return ret_val



"""
Various low level utilities
"""

import functools
import itertools
import re
import string
from typing import Optional, Any, List, Iterable


def is_empty(s):
    """
    True if None or string with whitespaces

    >>> is_empty(None)
    True
    >>> is_empty("hello")
    False
    >>> is_empty("  \t  ")
    True
    """
    return s is None or len(s) == 0 or s.isspace()


def _whitespace_regex():
    """
    >>> _whitespace_regex()
    '( |\\t|\\n|\\r|\\x0b|\\x0c)+'
    """
    all_white_spaces = [f'{x}' for x in string.whitespace]
    regex_string = functools.reduce(lambda acc, x: f"{acc}|{x}", all_white_spaces)
    return f"({regex_string})+"


_whitespace = re.compile(_whitespace_regex())


def normalize_string(val: Optional[str]):
    """
    Creates a string that is stripped of any white spaces at beg and end and
    has all sequences of whitespaces reduced to one single space

    >>> normalize_string('   this   is    \\t a very  \\n \\r spacy   string ')
    ' this is a very spacy string '
    """
    if val is None:
        return ""

    return _whitespace.sub(' ', val)


def flatten(val: Iterable[Iterable[Any]]) -> List[Any]:
    """
    Flatens a list of list into a list

    >>> flatten( [['abc','def'],[12,34,46],[3.14, 2.22]])
    ['abc', 'def', 12, 34, 46, 3.14, 2.22]
    """
    return list(itertools.chain(*val))



"""
Utilities for applying processors to files
"""
import shutil
from fnmatch import fnmatch
from typing import Callable, List, Any, Sequence
from bs4 import BeautifulSoup
import os
from os import path, walk
import logging
import re

_log = logging.getLogger("bs-processors")


def process_directory(processor: Callable[[List[Any]], List[Any]], parser_type: str,
                      input_dir: str, output_dir: str,
                      file_selector):
    """
    Processes a directory with the specified processor

    * **processor**: a file processor
    * **parser_type**: processor 'html.parser', 'html', 'xml' ( BeautifulSoup parser)
    * **input_dir**: the input directory
    * **output_dir**: the output directory
    * **file_selector**: something that can be transformed into a file_name predicate
        if the predicate is true than the file will be processed if not the file will be
        copied from input dir to output dir, see `to_file_selector_predicate` for details
        about the file selector.
    """
    file_selector = to_file_selector_predicate(file_selector)
    for dirpath, dirnames, filenames in walk(input_dir):
        rel_path = dirpath[len(input_dir):]
        if len(rel_path) > 0 and rel_path[0] == path.sep:
            rel_path= rel_path[1:]  # remove start '/'

        current_output_dir = path.join(output_dir, rel_path)

        if not path.exists(current_output_dir):
            os.makedirs(current_output_dir)

        for fname in filenames:
            input_fname = path.join(dirpath, fname)
            output_fname = path.join(current_output_dir, fname)
            if file_selector(input_fname):
                _log.debug(f"processing '{input_fname}' into '{output_fname}'")
                process_file(processor, parser_type, input_fname, output_fname)
            else:
                _log.debug(f"copying '{input_fname}' into '{output_fname}'")
                shutil.copy(input_fname, output_fname)


def process_file(processor: Callable[[List[Any]], List[Any]], parser_type: str, input_file: str, output_file: str):
    """
    Processes a file with the passed processor and saves the result in the output file

    * **processor**: the processor to be applied
    * **parser_type**: BeautifulSoup parser type ( 'html', 'xml', 'html.parser', etc)
    * **input_file**: the input file name
    * **output_file**: the result file name

    """
    with open(input_file, "rt") as f:
        soup = BeautifulSoup(f, parser_type)

    result = processor([soup])

    output_len = len(result)

    if output_len == 0:
        _log.warning(f"processing '{input_file}' did NOT generate any output")
        return

    if output_len > 1:
        _log.warning(f"processing '{input_file}' generated multiple output elements saving only the first one")

    result = result[0]

    if result.name != '[document]':
        _log.warning(f"processing '{input_file}' did not yield a beautiful soup element creating one")
        soup = BeautifulSoup(features=parser_type)
        result = soup.append(result)

    directory_name, f_name = path.split(output_file)

    if not path.exists(directory_name):
        os.makedirs(directory_name)

    with open(output_file, "wt") as f:
        f.write(result.prettify())


def process_html_file(processor: Callable[[List[Any]], List[Any]], input_file: str, output_file: str):
    process_file(processor, 'html.parser', input_file, output_file)


def to_file_selector_predicate(pred):
    """
    Creates a file selector predicate from a variety of arguments

    * **pred**: something that can be transformed in a file name predicate
         * None: will match everything
         * a str: will be interpreted as a unix file pattern (e.g. *.txt )
         * a sequence: will be interpreted as a sequence of unix file patterns
                (e.g. [*.txt, *.py]
         * a regular expression, will create a predicate with re.fullmath (i.e. full match
            of the full file name)
         * a predicate that takes a string (the full file name)

    * **return**: a file name predicate

    >>> pm = to_file_selector_predicate('*.txt')
    >>> pm('abc/def.txt')
    True
    >>> pm('abc/def.doc')
    False
    >>> pm = to_file_selector_predicate(['*.txt', '*.doc'])
    >>> pm('abc/def.doc')
    True
    >>> pm('abc/def.txt')
    True
    >>> pm('abc/def.tt')
    False
    >>> pm = to_file_selector_predicate(re.compile("(abc)|(def+)"))
    >>> pm("abc")
    True
    >>> pm("abcd")
    False
    >>> pm("def")
    True
    >>> pm("deffff")
    True
    >>> pm("something")
    False
    >>> pm = to_file_selector_predicate(lambda x: x.endswith("txt"))
    >>> pm("abc.txt")
    True
    >>> pm("abc.tt")
    False

    """
    if pred is None:
        return True  # select everything

    # pred is a string
    if isinstance(pred, str):
        return pattern_match_pred([pred])

    # pred is a list like object
    elif isinstance(pred, (tuple, list, set, frozenset)):
        return pattern_match_pred(pred)
    # pred is a regex
    elif isinstance(pred, re.Pattern):
        return lambda fname: pred.fullmatch(fname) is not None
    # pred must be a predicate, use it as is
    else:
        return pred


def pattern_match_pred(patterns: Sequence[str]) -> Callable[[str], bool]:
    """
    Creates a unix file pattern match predicate from a sequence of patterns

    * **patterns**: sequence of patterns
    * **return**: predicate

    >>> pm = pattern_match_pred(["*.exe", "*.txt", "*.do?"])
    >>> pm( "abc.txt")
    True
    >>> pm( "User/bubu/xyz.txt")
    True
    >>> pm( "abc.txta")
    False
    >>> pm('abc.exe')
    True
    >>> pm('abc.ex')
    False
    >>> pm('abc.doc')
    True
    """
    def inner(file_name: str) -> bool:
        for pattern in patterns:
            if fnmatch(file_name, pattern):
                return True
        return False

    return inner



"""
Generic modification factories.

This module contains factories that can be used to easily construct modification functions that can be
used as arguments to the local_modify function

"""
from typing import Any, Callable

from bs_processors import is_tag


def modify_if_f(modifier: Callable[[Any], None], predicate: Callable[[Any], bool]) -> Callable[[Any], None]:
    """
    Adaptor for a local modifier, it will be called only if the predicate returns True

    This modifier permits writing simple modifiers that apply their modification logic to all elements
    that they receive and delegate the selection logic to the predicate

    If, for example, we need a modifier that sets a certain class to all elements of type `p` we can create
    a modifier that sets the class to all elements that it receives and then use this function togehter with
    the `has_name_pf('p') to obtain the same result.

    :param modifier: a modifier that applies the modification to all elements it is called with
    :param predicate: a predicate that selects the elements to be modified
    :return: a modifier that applies the modification to the elements selected by the predicate

    >>> from bs4 import BeautifulSoup
    >>> from bs_processors.predicate import has_name_pf
    >>> from bs_processors.generic_processors import local_modify_factory
    >>> doc = '<html><span>s1</span><span class="c1">s2</span><div><span>s3</span></div></html>'
    >>> s = BeautifulSoup(doc, 'html.parser')
    >>> is_span = has_name_pf('span')
    >>> def modifier(elm):
    ...     clss = elm.get('class')
    ...     if clss is not None:
    ...         clss.append("my_class")
    ...     else:
    ...         elm['class'] = 'my_class'
    >>> modif_span = modify_if_f( modifier, is_span)
    >>> processor = local_modify_factory(modif_span)
    >>> result = processor([s])
    >>> result[0]
    <html><span class="my_class">s1</span><span class="c1 my_class">s2</span><div><span \
class="my_class">s3</span></div></html>
    """

    def inner(element):
        if predicate(element):
            modifier(element)


    return inner


def toggle_class(class_name_gen):
    """
    Toggles a class on a HTML element

    **Important**: this works with the assumption that the element is processed with 'html.parser' or something
    similar, that is the 'class' attribute is a list of classes. If processing with the 'html' parser the 'class'
    attribute is treated as a simple string and toggle_class will not function correctly.

    * **class_name_gen**: something that can be converted into a class name
        * st4 the string is the class name
        * Callable[[Any], str] a function that receives an element and returns a string, the class name to be toggled
    * **return**: a local modification function ( a function that can be applied on an element)

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><div></div><span id="1" class="s1 s2"/><span><p  class="a1"/><p></html>'
    >>> x = BeautifulSoup(doc, 'html.parser')
    >>> tc = toggle_class('bubu')
    >>> tc(x.div)
    >>> x.div
    <div class="bubu"></div>
    >>> tc(x.div)
    >>> x.div
    <div class=""></div>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2 bubu" id="1"></span>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2" id="1"></span>
    >>> tc(x.p)
    >>> x.p
    <p class="a1 bubu"></p>
    >>> tc(x.p)
    >>> x.p
    <p class="a1"></p>
    >>> class_name_gen = lambda x: x.name
    >>> tc = toggle_class(class_name_gen)
    >>> tc(x.div)
    >>> x.div
    <div class="div"></div>
    >>> tc(x.div)
    >>> x.div
    <div class=""></div>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2 span" id="1"></span>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2" id="1"></span>

    """
    class_name_gen = to_string_gen(class_name_gen)
    def inner(elm):
        if not is_tag(elm):
            return

        clss = elm.get("class")
        class_name = class_name_gen(elm)
        if clss is None:
            elm['class'] = [class_name]
        else:
            if class_name in clss:
                elm['class']= [cls for cls in clss if cls != class_name]
            else:
                clss.append(class_name)

    return inner

def remove_class(class_name_gen):
    """
    Removes a class on a HTML element

    **Important**: this works with the assumption that the element is processed with 'html.parser' or something
    similar, that is the 'class' attribute is a list of classes. If processing with the 'html' parser the 'class'
    attribute is treated as a simple string and toggle_class will not function correctly.

    * **class_name_gen**: something that can be converted into a class name
        * st4 the string is the class name
        * Callable[[Any], str] a function that receives an element and returns a string, the class name to be toggled
    * **return**: a local modification function ( a function that can be applied on an element)

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><div></div><span id="1" class="s1 s2"/><span></html>'
    >>> x = BeautifulSoup(doc, 'html.parser')
    >>> tc = remove_class('s1')
    >>> tc(x.div)
    >>> x.div
    <div></div>
    >>> tc(x.span)
    >>> x.span
    <span class="s2" id="1"></span>
    """
    class_name_gen = to_string_gen(class_name_gen)
    def inner(elm):
        if not is_tag(elm):
            return

        clss = elm.get("class")
        class_name = class_name_gen(elm)
        if clss is not None:
            elm['class']= [cls for cls in clss if cls != class_name]

    return inner

def set_class(class_name_gen):
    """
    Sets a class (does not duplicate the class if it already exists).

    Unlike add_class it will not add a class multiple times (if the class is already set it will not add it again)

    **Important**: this works with the assumption that the element is processed with 'html.parser' or something
    similar, that is the 'class' attribute is a list of classes. If processing with the 'html' parser the 'class'
    attribute is treated as a simple string and toggle_class will not function correctly.

     * **class_name_gen**: something that can be converted into a class name
        * st4 the string is the class name
        * Callable[[Any], str] a function that receives an element and returns a string, the class name to be toggled
    * **return**: local modification function

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><div></div><span id="1" class="s1 s2"/><span></html>'
    >>> x = BeautifulSoup(doc, 'html.parser')
    >>> tc = set_class('s1')
    >>> tc(x.div)
    >>> x.div
    <div class="s1"></div>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2" id="1"></span>
    """
    class_name_gen = to_string_gen(class_name_gen)
    def inner(elm):
        if not is_tag(elm):
            return

        clss = elm.get("class")
        class_name = class_name_gen(elm)
        if clss is None:
            elm['class'] = [class_name]
        else:
            if class_name not in clss:
                clss.append(class_name)

    return inner

def add_class(class_name_gen):
    """
    Toggles a class on a HTML element

    Unlike set_class it will add a class multiple times (if the class is already set it will add it again)

    **Important**: this works with the assumption that the element is processed with 'html.parser' or something
    similar, that is the 'class' attribute is a list of classes. If processing with the 'html' parser the 'class'
    attribute is treated as a simple string and toggle_class will not function correctly.

    * **class_name_gen**: something that can be converted into a class name
        * st4 the string is the class name
        * Callable[[Any], str] a function that receives an element and returns a string, the class name to be toggled
    * **return**: a local modification function ( a function that can be applied on an element)

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><div></div><span id="1" class="s1 s2"/><span></html>'
    >>> x = BeautifulSoup(doc, 'html.parser')
    >>> tc = add_class('s1')
    >>> tc(x.div)
    >>> x.div
    <div class="s1"></div>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2 s1" id="1"></span>
    >>> tc(x.span)
    >>> x.span
    <span class="s1 s2 s1 s1" id="1"></span>
    """

    class_name_gen = to_string_gen(class_name_gen)
    def inner(elm):
        if not is_tag(elm):
            return

        clss = elm.get("class")
        class_name = class_name_gen(elm)
        if clss is None:
            elm['class'] = [class_name]
        else:
            clss.append(class_name)

    return inner


def change_tag_name(tag_name_gen):
    """
    Changes the tag name of a tag (e.g. change `div` to `p`)


    * **tag_name_gen**: something that can be converted into a class name
        * st4 the string is the new tag name
        * Callable[[Any], str] a function that receives an element and returns a string, the new tag name to be set
    * **return**: a local modification function ( a function that can be applied on an element)

    >>> from bs4 import BeautifulSoup
    >>> doc = '<html><div></div><span id="1" class="s1 s2"></span></html>'
    >>> x = BeautifulSoup(doc, 'html.parser')
    >>> tc = change_tag_name('p')
    >>> tc(x.div)
    >>> x
    <html><p></p><span class="s1 s2" id="1"></span></html>
    >>> tc(x.span)
    >>> x
    <html><p></p><p class="s1 s2" id="1"></p></html>
    """
    tag_name_gen = to_string_gen(tag_name_gen)
    def inner(elm):
        if not is_tag(elm):
            return

        elm.name = tag_name_gen(elm)

    return inner


def to_string_gen(gen) -> Callable[[Any], str]:
    """
    Accepts a string or a function that returns a string and returns a function that returns a string

    :param gen: either the string to be returned or a generator that accepts an element and returns a string
    :return:  a string generator that accepts an element

    >>> s_gen = to_string_gen("some_constant")
    >>> s_gen(1)
    'some_constant'
    >>> s_gen("hello")
    'some_constant'
    >>> echo_gen = to_string_gen( lambda x: str(x))
    >>> echo_gen(1)
    '1'
    >>> echo_gen("hello")
    'hello'
    """
    if gen is None:
        return lambda x: ""
    if isinstance(gen, str):
        return lambda x: gen
    else:
        return gen




"""
Generic processor factories

This module contains basic types of processor factories that can be used to
generate typical processors ( like filter, flatten, local_modify ... )
"""

import functools
from typing import Callable, Any, List, Sequence

from .processor_util import single_to_multiple
from .xml_util import set_new_children, process_children, copy_element_type

# configure the documented entities
__pdoc__={}


def single_filter_proc(should_filter: Callable[[Any], bool], elm) -> List[Any]:
    """
    Conditionally filters an element based on the passed predicate

    * **should_filter**: predicate that returns True if the element should be filtered
    * **elm**: the element to check
    * **return**: an empty array if the element should be filtered or an array with the passed element

    >>> def should_filter(elm):
    ...     if elm.name in ['span', 'a']:
    ...         return True
    ...     return False
    ...
    >>> from bs4 import BeautifulSoup as bs
    >>> doc = bs('<html><div>a<span>b</span> <a></div> <p><span>x</span></p><a></html>','html.parser')
    >>> filtered = single_filter_proc(should_filter, doc.html)
    >>> filtered
    [<html><div>a </div> <p></p></html>]

    """
    if should_filter(elm):
        return []
    else:
        new_children = process_children(lambda child: single_filter_proc(should_filter, child), elm)
        set_new_children(elm, new_children)

        return [elm]


def filter_factory(should_filter):
    """
    Adaptor for single_filter_proc to accept multiple elements
    """
    return single_to_multiple(functools.partial(single_filter_proc, should_filter))


def single_unwrap_proc(should_unwrap: Callable[[Any], bool], elm) -> List[Any]:
    """
    Conditionally unwraps an element and pushes all its unwrapped children to its parent

    * **should_unwrap**: predicate that returns true if the element should be unwrapped
    * **elm**: the element to be conditionally unwrapped
    * **return**: either a list containing the element (if it wasn't unwrapped) or a list with all
     the unwrapped children

    >>> from bs4 import BeautifulSoup
    >>> def should_unwrap(elm):
    ...     return elm.name == "x"
    ...

    >>> doc = "<root><x >hello</x> after x <b> in b</b></root>"
    >>> elm = BeautifulSoup(doc, "xml")
    >>> # single_unwrap_proc(should_unwrap, elm.root)
    [<root>hello after x <b> in b</b></root>]

    >>> doc = "<root><x >hello <a> in a</a></x></root>"
    >>> elm = BeautifulSoup(doc, "xml")
    >>> single_unwrap_proc(should_unwrap, elm.root)
    [<root>hello <a> in a</a></root>]

    >>> doc = "<root><x >hello <a> in a <x>in x2</x> after</a></x> 22 <b>in b</b> end</root>"
    >>> elm = BeautifulSoup(doc, "xml")
    >>> single_unwrap_proc(should_unwrap, elm.root)
    [<root>hello <a> in a in x2 after</a> 22 <b>in b</b> end</root>]
    """

    # Note: elm.unwrap() would do the job but the return value is not
    # what I need (it is the node that was unwrapped instead of the
    # children) so we need to do it manually
    new_children = process_children(lambda child: single_unwrap_proc(should_unwrap, child), elm)

    if not should_unwrap(elm):
        set_new_children(elm, new_children)
        return [elm]

    # insert the new children in the parent
    # TODO see if I need to do it this way !!!
    current = elm
    for child in new_children:
        current.insert_after(child)
        current = child
    elm.extract()  # remove current element
    return new_children


def unwrap_factory(should_unwrap):
    """
    Adaptor for single_unwrap_proc to accept multiple elements
    """
    return single_to_multiple(functools.partial(single_unwrap_proc, should_unwrap))


def single_flatten_proc(flatten_children: Callable[[Any], bool], is_internal: Callable[[Any], bool], elm) -> List[Any]:
    """
    Flattens an element pulling out 'block' like elements that 'want' to be top level

    * **flatten_children**: Predicate that specifies whether the direct children of the current element can be
        flatten (i.e. taken outside of this element) (e.g. html, body and a do not allow their children to escape
        outside)
    * **is_internal**: Predicate that specifies whether a child 'wants' to stay inside its parent or not (i.e.
        whether a child behaves like a block and wants out or an inline and stays in)
    * **elm**: the element to be flatten
    * **return**: a list of flattened elements

    >>> from bs4 import BeautifulSoup
    >>> def flatten_children(elm):
    ...     if elm.name is None or elm.name in {'a', 'body', 'html'}:
    ...         return False
    ...     return True
    ...
    >>> def is_internal(elm):
    ...     if elm.name is None:
    ...         return True
    ...     if elm.name in { 'div', 'p', 'br'}:
    ...         return False
    ...     if elm.name == 'a' and "block_a" in elm['class']:
    ...         return False
    ...     return True
    ...

    >>> doc = '<div id="1"> a <br id="2"> b</div>'
    >>> s = BeautifulSoup(doc, "html.parser")
    >>> elm = s.div
    >>> single_flatten_proc(flatten_children, is_internal, elm)
    [<div id="1"> a </div>, <br id="2"/>, <div> b</div>]

    >>> doc = '<div id="1"><i>it</i><br id="2"><b></b></div>'
    >>> s = BeautifulSoup(doc, "html.parser")
    >>> elm = s.div
    >>> single_flatten_proc(flatten_children, is_internal, elm)
    [<div id="1"><i>it</i></div>, <br id="2"/>, <div><b></b></div>]

    >>> doc = '<div id="1">a<a class="block_a">b</a>c</div>'
    >>> s = BeautifulSoup(doc, "html.parser")
    >>> elm = s.div
    >>> single_flatten_proc(flatten_children, is_internal, elm)
    [<div id="1">a</div>, <a class="block_a">b</a>, <div>c</div>]

    >>> doc = '<div id="1"><a class="block_a"><div>inside a</div></a></div>'
    >>> s = BeautifulSoup(doc, "html.parser")
    >>> elm = s.div
    >>> single_flatten_proc(flatten_children, is_internal, elm)
    [<div id="1"></div>, <a class="block_a"><div>inside a</div></a>]
    """
    new_children = process_children(lambda child: single_flatten_proc(flatten_children, is_internal, child), elm)

    if not flatten_children(elm):
        # all children should go inside
        set_new_children(elm, new_children)
        return [elm]

    elm.clear()
    # this element pops out external children
    # remember the current parent
    parent = elm

    result = []
    for child in new_children:
        if is_internal(child):
            if parent is None:
                # we need a new element like elm
                parent = copy_element_type(elm)
            parent.append(child)
        else:
            # we need to pop the element at a higher level, if we have a valid parent
            # append it first
            if parent is not None:
                result.append(parent)
                parent = None
            result.append(child)

    if parent is not None:
        result.append(parent)

    return result


def flatten_factory(flatten_children, is_internal):
    """
    Adaptor for single_filter_proc to accept multiple elements
    """
    return single_to_multiple(functools.partial(single_filter_proc, should_filter))

    return single_to_multiple(functools.partial(single_flatten_proc, flatten_children, is_internal))


def single_local_modify(modify_func: Callable[[Any], None], elm):
    """
    Runs a modification function on each element and all tag children

    * **modify_func**: the modification function
    * **elm**: the top level element
    * **return**: an array containing the original element
    """
    modify_func(elm)
    process_children(lambda child: single_local_modify(modify_func, child), elm)
    return [elm]


def local_modify_factory(modify_func):
    """
    Adaptor for single_local_modify to accept multiple elements
    """
    return single_to_multiple(functools.partial(single_local_modify, modify_func))


def single_join_children(join_children: Callable[[Any, Any], Any], elm)-> List[Any]:
    """
    Processor that joins children

    * **join_children**: a function that may join two children. If the children are
    joined the function should return an array containing the joined element, if the
    children are not joined the function should return an array containing the passed
    children.
    * **elm**: the root element
    * **return**: a sequence of elements
    """
    new_children = process_children( join_children, elm)

    def reducer(accumulator, new_child):
        if len(accumulator) > 0:
            # try to see if we can join the last child with the new child
            result = accumulator[:-1] + list(join_children(accumulator[-1], new_child))
        else:
            # this is the first element just store it (noting to join it to)
            result = [new_child]
        return result

    # try to join adjacent children
    joined_children = functools.reduce(reducer, new_children, [])
    set_new_children(elm, joined_children)
    return [elm]


def join_children_factory(join_children):
    """
    Adaptor for single_join_children to accept multiple elements
    """
    return single_to_multiple(functools.partial(single_join_children, join_children))


def lateral_effect(lateral_effect_func, elms: Sequence[Any]) -> Sequence[Any]:
    lateral_effect_func()
    return elms


__pdoc__["lateral_effect"] = False  # do not document 


def lateral_effect_factory(lateral_effect_func):
    """
    A processor factory that runs a lateral effect ( provided function) and the returns the
    passed argument
    """
    return functools.partial(lateral_effect, lateral_effect_func)





import pytest
from os import path, remove
from lxml import etree


@pytest.fixture
def path_resolver():
    def inner(*args):
        return path.abspath(path.join(*args))

    return inner


@pytest.fixture
def html_file_loader():
    def inner(path):
        with open(path, 'r') as f:
            parser = etree.HTMLParser()
            tree = etree.parse(f, parser)
            return tree.getroot()

    return inner


@pytest.fixture
def dump_xml_file():
    def inner(path, elm):
        with open(path, "wb") as f:
            tree = etree.ElementTree(elm)
            tree.write(f, pretty_print=True)

    return inner


@pytest.fixture
def result_file_logger(request):
    params = {}

    def inner(file_name, file_generator, *args, **kwargs):
        # if we don't yet have a result create one
        if not path.exists(file_name):
            file_generator(file_name, *args, **kwargs)

        dir, f_name = path.split(file_name)
        components = f_name.split('.', 1)
        if len(components) > 1:
            f_name = components[0] + "_." + components[1]
        else:
            f_name += "_"
        error_file_name = path.join(dir, f_name)

        if path.exists(error_file_name):
            # we have an old error result, delete it
            remove(error_file_name)
        params['error_file_name'] = error_file_name
        params['file_generator'] = file_generator
        params['args'] = args
        params['kwargs'] = kwargs

    yield inner
    if request.node.rep_setup.passed and request.node.rep_call.failed:
        # we have a failed test
        fun = params.get('file_generator')
        if fun is not None:
            args = params['args']
            kwargs = params['kwargs']
            error_file_name = params['error_file_name']
            fun(error_file_name, *args, **kwargs)



"""
Generic utilities for processors

Contains utilities for assembling and modifying processors
"""
from typing import Sequence, Callable, Any, List


def join(processors: Sequence[Callable[[List[Any]], List[Any]]])-> Callable[[List[Any]], List[Any]]:
    """
    Creates a processor from a list of processor.
    The processor created is equivalent to applying each of the passed processor
    in sequence over the result of the previous processor (after the result is
    flattened)

    * **processors**: a list of processors
    * **return** : a join processor

    >>> def double( elms):
    ...     result = []
    ...     for elm in elms:
    ...         result += [elm, elm]
    ...     return result
    ...
    >>> def add_one( elms):
    ...     result = []
    ...     for elm in elms:
    ...         result.append(elm + 1)
    ...     return result
    ...
    >>> pr = join([add_one, double])
    >>> pr( [1,100])
    [2, 2, 101, 101]
    >>> pr = join([double, add_one])
    >>> pr([1, 100])
    [2, 2, 101, 101]
    >>> pr = join([double, add_one, double])
    >>> pr([1, 100])
    [2, 2, 2, 2, 101, 101, 101, 101]
    """
    def inner(elms):
        result = elms
        for processor in processors:
            result = processor(result)
        return result
    return inner


def single_to_multiple(processor: Callable[[Any], List[Any]])->Callable[[List[Any]], List[Any]]:
    """
    Creates a processor that accepts a list of elements.

    Converts a processor that handles one element and returns a list of elements into
    a processor that handles a list of elements and returns a list of elements by calling
    each element in the list in turn and joining the intermediate results into one result

    * **processor**: a processor that operates on one element
    * **return** : a processor that operates on a list of elements

    >>> def square_pair(x):
    ...     return[x , x*x]
    >>> mult = single_to_multiple(square_pair)
    >>> mult([1,2,3,4])
    [1, 1, 2, 4, 3, 9, 4, 16]
    """

    def inner( elms: List[Any])-> List[Any]:
        result = []
        for elm in elms:
            if elm is not None:
                result += processor(elm)
        return result

    return inner

def process_file(processor: Callable[[Any], List[Any]], input_fname: str, output_fname: str):
    pass



"""
Utilities for working with BeautifulSoup trees

This module contains utilities useful for implementing custom processors and predicates.
"""
from typing import Any, Callable, List

from bs4 import BeautifulSoup


def set_new_children(elm, children):
    """
    Sets new children to an element

    * **elm**: the element to have its children set
    * **children**: a sequence of children
    * **return**: the element

    >>> xml = BeautifulSoup("<root>X<a/>X<b/>X</root>",'xml')
    >>> xml
    <?xml version="1.0" encoding="utf-8"?>
    <root>X<a/>X<b/>X</root>
    >>> new_children = [xml.new_tag("x"), xml.new_tag("y")]
    >>> set_new_children( xml.root, new_children)
    <root><x/><y/></root>

    """

    new_children = list(children)
    if list(elm.children) == new_children:
        return elm

    elm.clear()
    for child in new_children:
        elm.append(child)

    return elm


def process_children(processor: Callable[[Any], List[Any]], elm):
    """
    Processes the children of an element

    * **processor**: A processor that takes an element and returns a list of elements
    * **elm**: the element to process
    * **return**: A list with the joined results of processing each child with the provided processor

    >>> doc = BeautifulSoup("<div>a<span>b</span>c<p>d<span>e</span>f<a/>g</p>h</div>", "html.parser")
    >>> counter = [0]
    >>> def processor(elm):
    ...     counter[0] +=1
    ...     elm['id'] = str(counter[0])
    ...     return [elm]
    >>> process_children(processor, doc.div)
    ['a', <span id="1">b</span>, 'c', <p id="2">d<span>e</span>f<a></a>g</p>, 'h']
    """
    result = []
    for child in elm.children:
        if is_tag(child):
            result += processor(child)
        else:
            result.append(child)
    return result


def is_tag(elm):
    """
    True if the passed object is a HTML/XML tag

    * **elm**:
    * **return**:

    >>> from bs4 import BeautifulSoup as bs
    >>> doc = bs("<span></span><p>bubu</p>", "html.parser")
    >>> span = doc.span
    >>> is_tag(doc.span)
    True
    >>> is_tag(doc.p)
    True
    >>> is_tag(list(doc.p.children)[0])
    False
    """
    return elm is not None and elm.name is not None


_bs = BeautifulSoup("<a/>", "html.parser")


def copy_element_type(elm):
    """
    Creates a new element with the same tag name as the passed element

    >>> xml = BeautifulSoup("<root><abc/></root>", "xml")
    >>> copy_element_type(xml.root.abc)
    <abc></abc>
    """
    return _bs.new_tag(elm.name)



import click



@click.group()
def main():
    pass


@click.option("--output-dir", prompt="output directory", help="the input directory")
@click.option("--input-dir", prompt="input directory", help="the input directory")
@click.option("--processor", prompt="processor file", help="the processor file")
@main.command()
def p_dir(input_dir, output_dir, processor):
    click.echo(f"hello {input_dir} from main")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bs-processors",
    version="0.0.1pre10",
    author="RaduW",
    description="html/xml processors for using with BeautifulSoup",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://raduw.github.io/bs-processors/",
    entry_points = {
      "console_scripts": ['bs-proc=bs_processors.cli.main:main'],
    },
    project_urls = {
        "Documentation": "https://raduw.github.io/bs-processors/",
        "Source Code": "https://github.com/RaduW/bs-processors.git",
        "Bug Tracker": "https://github.com/RaduW/bs-processors/issues",
    },
    install_requires=[
        "click", "beautifulsoup4"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries"
    ],
    keywords ="BeautifulSoup",
    python_requires=">=3.7",
    license="MIT",
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
)



"""
Xml comparison utility.

"""
from bs_processors.utils.util import normalize_string, is_empty


def _not_empty(elm):
    """
    duplicates not_empty_p functionality so that it doesn't import predicate
    """
    if elm is None:
        return False
    if elm.name is not None:
        return True
    return not is_empty(elm)



def compare_xml(left, right, error_context = None):
    """
    Compares two elements
    * **left**: the left element
    * **right**: the right element
    * **error_context**: error context should be None at top level
    * **return**: None if contents are identical, an error message if they are not

    >>> from bs4 import BeautifulSoup as bs
    >>> doc1 = bs("<html> hello  </html>", "html.parser")
    >>> doc2 = bs("<html>hello</html>", "html.parser")
    >>> compare_xml(doc1.html, doc2.html) is None
    True

    >>> doc1 = bs("<html> <p>x</p>  </html>", "html.parser")
    >>> doc2 = bs("<html><p>x</p</html>", "html.parser")
    >>> compare_xml(doc1.html, doc2.html) is None
    True

    >>> doc1 = bs("<html><p>x</p><p>x</p></html>", "html.parser")
    >>> doc2 = bs("<html><p>x</p></html>", "html.parser")
    >>> compare_xml(doc1.html, doc2.html)
    'Different number of children at:html'

    >>> doc1 = bs("<html><i>x</i></html>", "html.parser")
    >>> doc2 = bs("<html><p>x</p</html>", "html.parser")
    >>> compare_xml(doc1.html, doc2.html)
    'Different tag name at: html->[0]i'

    >>> doc1 = bs("<html> <p id='1'>x</p>  </html>", "html.parser")
    >>> doc2 = bs('<html><p id="1" class="x">x</p</html>', "html.parser")
    >>> compare_xml(doc1.html, doc2.html)
    'Different attributes at: html->[0]p#1'

    >>> doc1 = bs("<html> hello  </html>", "html.parser")
    >>> doc2 = bs("<html>hello2</html>", "html.parser")
    >>> compare_xml(doc1.html, doc2.html)
    "different string content 'hello'!='hello2' at html->[0]None"
    """
    if error_context is None:
        error_context = ""
    error_context= "{}{}".format(error_context,left.name)

    if left is None and right is None:
        return None

    if left is None :
        return "None object to the left at :{}".format(error_context)
    if right is None:
        return "None object to the right at :{}".format(error_context)

    # check strings
    if left.name is None and right.name is None:
        text_left = normalize_string(left).strip()
        text_right = normalize_string(right).strip()
        if text_left != text_right:
            return f"different string content '{text_left}'!='{text_right}' at {error_context}"
        else:
            return None
    if left.name is None:
        return f"left is string while right is elm left->'{left}' right->'{right.name}' at {error_context}"
    if right.name is None:
        return f"left is elm while right is string left->'{left.name}' right->'{right}' at {error_context}"

    elm_id = left.get('id')
    elm_cls = left.get('class')
    if elm_id is not None:
        error_context = "{}#{}".format(error_context,elm_id)
    elif elm_cls is not None:
        # we don't have an id but we have some class
        error_context = "{}<{}>".format(error_context,elm_cls)

    if left.name != right.name:
        return "Different tag name at: {}".format(error_context)

    if left.attrs != right.attrs:
        return "Different attributes at: {}".format(error_context)

    # remove text children with empty spaces (i.e. format differences) from comparison
    children_left = list(filter(_not_empty, left.children))
    children_right = list(filter(_not_empty, right.children))

    if len(children_left) != len(children_right):
        return "Different number of children at:{}".format(error_context)

    for idx in range(len(children_left)):
        new_context= "{}->[{}]".format(error_context, idx)
        result = compare_xml(children_left[idx], children_right[idx], new_context)
        if result is not None:
            return result
    return None  # success



from .generic_processors import (
    filter_factory, unwrap_factory, flatten_factory, local_modify_factory,
    join_children_factory, lateral_effect_factory,
)
from .processor_util import join
from .predicate import *

__pdoc__={}
__pdoc__["bs_processors.utils.pytest"] = False  # hide from doc
